"""
# Spectre-Format Netlister
"""

# Std-Lib Imports
from typing import Union, List

# Local Imports
import vlsir

# Import the base-class
from .base import Netlister, ResolvedModule, ResolvedParams, SpicePrefix


def map_primitive(rmodule: ResolvedModule, paramvals: ResolvedParams) -> str:
    """ Map a primitive into Spectre's supported names and parameters. 
    Returns the "apparent module name" for instances of the primitive. 
    Argument `paramvals` is often modified along the way. 

    Note spectre syntax is such that the "apparent module name" can be either of: 
    (a) a fixed name per spice-prefix, for basic types (r, c, l, etc), or 
    (b) the *model* name for model-based devices, (mos, bjt, diode, etc) """

    # For voltage sources, add spectre's "type" parameter, and potentially rename several parameters
    if rmodule.spice_prefix == SpicePrefix.VSOURCE:
        vname = rmodule.module_name
        vtypes = dict(vdc="dc", vpulse="pulse", vsin="sine",)
        if vname not in vtypes:
            msg = f"Invalid or unsupported voltage-source type {vname}"
            raise ValueError(msg)
        paramvals.set("type", vtypes[vname])

        if vname == "vpulse":
            # For pulse sources, need to rename most parameters
            paramvals.rename("v1", "val0")
            paramvals.rename("v2", "val1")
            paramvals.rename("td", "delay")
            paramvals.rename("tr", "rise")
            paramvals.rename("tf", "fall")
            paramvals.rename("tpw", "width")
            paramvals.rename("tper", "period")

    # Mapping from spice-prefix to spectre-name for fixed-name types
    basics = {
        SpicePrefix.RESISTOR: "resistor",
        SpicePrefix.CAPACITOR: "capacitor",
        SpicePrefix.INDUCTOR: "inductor",
        SpicePrefix.VSOURCE: "vsource",
        SpicePrefix.ISOURCE: "isource",
        SpicePrefix.VCVS: "vcvs",
        SpicePrefix.VCCS: "vccs",
        SpicePrefix.CCCS: "cccs",
        SpicePrefix.CCVS: "ccvs",
    }
    if rmodule.spice_prefix in basics:
        return basics[rmodule.spice_prefix]

    # Other primitive-types get an "apparent module name" equal to their *model* name.
    model_based = {
        SpicePrefix.MOS,
        SpicePrefix.BIPOLAR,
        SpicePrefix.DIODE,
        SpicePrefix.TLINE,
    }
    if rmodule.spice_prefix in model_based:
        # Get the model-name from its instance parameters
        return paramvals.pop("modelname")

    # Otherwise, unclear what this is or how we got here.
    raise RuntimeError(f"Unsupported or Invalid Primitive {rmodule}")


class SpectreNetlister(Netlister):
    """ Spectre-Format Netlister """

    @property
    def enum(self):
        """ Get our entry in the `NetlistFormat` enumeration """
        from . import NetlistFormat

        return NetlistFormat.SPECTRE

    def write_module_definition(self, module: vlsir.circuit.Module) -> None:
        """ Create a Spectre-format definition for proto-Module `module` """

        # Create the module name
        module_name = self.get_module_name(module)
        # Check for double-definition
        if module_name in self.module_names:
            raise RuntimeError(f"Module {module_name} doubly defined")
        # Add to our visited lists
        self.module_names.add(module_name)
        self.pmodules[module.name] = module

        # Create the sub-circuit definition header
        self.write(f"subckt {module_name} \n")

        if module.ports:  # Create its ports
            self.write("+  ")
            for pport in module.ports:
                self.write(self.format_port_decl(pport) + " ")
            self.write("\n")
        else:
            self.write("+  // No ports \n")

        # Create its parameters, if defined
        if module.parameters:
            self.write("parameters ")
            for name, pparam in module.parameters.items():
                self.write(
                    self.format_param_decl(name, pparam)
                )  # FIXME! NotImplemented
            self.write("\n")
        else:
            self.write("+  // No parameters \n")

        self.write("\n")  # End "header" facets
        # Note nothing need be done for internal signals;
        # spice and spectre create these "out of thin air"

        # Create its instances
        for pinst in module.instances:
            self.write_instance(pinst)
        self.write("\n")

        # Close up the sub-circuit
        self.write("ends \n\n")

    def write_instance(self, pinst: vlsir.circuit.Instance) -> None:
        """Create and return a netlist-string for Instance `pinst`"""

        # Initial resolution phase.
        # Start by getting the Module or ExternalModule definition
        rmodule = self.resolve_reference(pinst.module)

        # Resolve its parameter values
        resolved_instance_parameters = self.get_instance_params(pinst, rmodule.module)

        module, module_name = rmodule.module, rmodule.module_name
        if rmodule.spice_prefix == SpicePrefix.SUBCKT:
            module_name = rmodule.module_name
        else:  # Primitive element. Look up spectre-format module-name
            module_name = map_primitive(rmodule, resolved_instance_parameters)

        # Create the instance name
        self.write(pinst.name + "\n")

        if module.ports:
            self.write("+  // Ports: \n")
            self.write("+  ( ")
            # Get `module`'s port-order
            port_order = [pport.signal.name for pport in module.ports]
            # And write the Instance ports, in that order
            for pname in port_order:
                pconn = pinst.connections.get(pname, None)
                if pconn is None:
                    raise RuntimeError(f"Unconnected Port {pname} on {pinst.name}")
                self.write(self.format_connection(pconn) + " ")
            self.write(" ) \n")
        else:
            self.write("+  // No ports \n")

        # Write the module-name
        self.write("+  " + module_name + " \n")

        if resolved_instance_parameters:  # Write its parameter-values
            self.write("+  ")
            for pname, pval in resolved_instance_parameters.items():
                self.write(f"{pname}={pval} ")
            self.write(" \n")
        else:
            self.write("+  // No parameters \n")

        # And add a post-instance blank line
        self.write("\n")

    def format_concat(self, pconc: vlsir.circuit.Concat) -> str:
        """ Format the Concatenation of several other Connections """
        out = ""
        for part in pconc.parts:
            out += self.format_connection(part) + " "
        return out

    @classmethod
    def format_port_decl(cls, pport: vlsir.circuit.Port) -> str:
        """ Get a netlist `Port` definition """
        return cls.format_signal_ref(pport.signal)

    @classmethod
    def format_port_ref(cls, pport: vlsir.circuit.Port) -> str:
        """ Get a netlist `Port` reference """
        return cls.format_signal_ref(pport.signal)

    @classmethod
    def format_signal_ref(cls, psig: vlsir.circuit.Signal) -> str:
        """ Get a netlist definition for Signal `psig` """
        if psig.width < 1:
            raise RuntimeError
        if psig.width == 1:  # width==1, i.e. a scalar signal
            return psig.name
        # Vector/ multi "bit" Signal. Creates several spice signals.
        return " ".join(
            [f"{psig.name}{cls.format_bus_bit(k)}" for k in reversed(range(psig.width))]
        )

    @classmethod
    def format_signal_slice(cls, pslice: vlsir.circuit.Slice) -> str:
        """Get a netlist definition for Signal-Slice `pslice`"""
        base = pslice.signal
        indices = list(reversed(range(pslice.bot, pslice.top + 1)))
        if not len(indices):
            raise RuntimeError(f"Attempting to netlist empty slice {pslice}")
        return " ".join([f"{base}{cls.format_bus_bit(k)}" for k in indices])

    @classmethod
    def format_bus_bit(cls, index: Union[int, str]) -> str:
        """ Format-specific string-representation of a bus bit-index"""
        # Spectre netlisting uses an underscore prefix, e.g. `bus_0`
        return "_" + str(index)

    def write_comment(self, comment: str) -> None:
        """ While Spectre *can* do a bunch of other comment-styles, 
        the canonical one is generally the C-style line comment beginning with `//`. """
        self.write(f"// {comment}\n")




"""
# Verilog-Format Netlister 
"""
# Local Imports
import vlsir

# Import the base-class
from .base import Netlister, SpicePrefix


class VerilogNetlister(Netlister):
    """
    # Structural Verilog Netlister 
    """

    @property
    def enum(self):
        """ Get our entry in the `NetlistFormat` enumeration """
        from . import NetlistFormat

        return NetlistFormat.VERILOG

    def write_module_definition(self, module: vlsir.circuit.Module) -> None:
        """ Create a Verilog module definition for proto-Module `module` """

        # Create the module name
        module_name = self.get_module_name(module)
        # Check for double-definition
        if module_name in self.module_names:
            raise RuntimeError(f"Module {module_name} doubly defined")
        # Add to our visited lists
        self.module_names.add(module_name)
        self.pmodules[module.name] = module

        # Create the module header
        self.writeln(f"module {module_name}")

        # Create its parameters, if defined
        if module.parameters:
            self.writeln("#( ")
            self.indent += 1
            for num, name in enumerate(module.parameters):
                pparam = module.parameters[name]
                comma = "" if num == len(module.parameters) - 1 else ","
                self.writeln(self.format_param_decl(name, pparam) + comma)
            self.indent -= 1
            self.writeln(") ")
        else:
            self.writeln("// No parameters ")

        if module.ports:  # Create its ports
            # Don't forget, a trailing comma after the last one is fatal to high-tech Verilog parsers!
            self.writeln("( ")
            self.indent += 1
            for num, pport in enumerate(module.ports):
                comma = "" if num == len(module.ports) - 1 else ","
                self.writeln(self.format_port_decl(pport) + comma)
            self.indent -= 1
            self.writeln("); ")
        else:
            self.writeln("// No ports ")

        self.writeln("")  # Blank line to end "header" facets
        self.indent += 1

        if module.signals:  # Create Signal declarations
            self.writeln("// Signal Declarations")
            for psig in module.signals:
                self.writeln(self.format_signal_decl(psig) + "; ")
        else:
            self.writeln("// No Signal Declarations")

        if module.instances:  # Create its instances
            self.writeln("")
            self.writeln("// Instance Declarations")
            for pinst in module.instances:
                self.write_instance(pinst)
        else:
            self.writeln("// No Instances")

        # Close up the module
        self.indent -= 1
        self.writeln("")  # Blank before `endmodule`
        self.writeln(f"endmodule // {module_name} \n\n")

    def write_instance(self, pinst: vlsir.circuit.Instance) -> None:
        """ Format and write Instance `pinst` """

        # Get its Module or ExternalModule definition
        rmodule = self.resolve_reference(pinst.module)
        if rmodule.spice_prefix != SpicePrefix.SUBCKT:
            # Spice-level primitives are generally not available in Verilog runtimes,
            # and hence generate errors here if attempted in netlisting.
            raise RuntimeError(f"Invalid module for Verilog: {rmodule}")
        module, module_name = rmodule.module, rmodule.module_name

        # Write the module-name
        self.writeln(module_name)

        if pinst.parameters:  # Write the parameter-values
            self.writeln("#( ")
            self.indent += 1
            for pname, pparam in pinst.parameters.items():
                pval = self.get_param_value(pparam)
                self.writeln(f"{pname}={pval} ")
            self.indent -= 1
            self.writeln(") ")
        else:
            self.writeln("// No parameters ")

        # Write the instance name
        self.writeln(pinst.name)

        if module.ports:  # Write connections, by-name, in-order
            self.writeln("( ")
            self.indent += 1
            # Get `module`'s port-order
            port_order = [pport.signal.name for pport in module.ports]
            # And write the Instance ports, in that order
            for num, pname in enumerate(port_order):
                pconn = pinst.connections.get(pname, None)
                if pconn is None:
                    raise RuntimeError(f"Unconnected Port {pname} on {pinst.name}")
                # Again a trailing comma after the last one is fatal!
                comma = "" if num == len(port_order) - 1 else ","
                self.writeln(f".{pname}({self.format_connection(pconn)}){comma} ")
            # Close up the ports
            self.indent -= 1
            self.writeln("); ")
        else:
            self.writeln("// No ports ")

        self.writeln("")  # Post-Instance blank line

    @classmethod
    def format_param_type(cls, pparam: vlsir.Param) -> str:
        """ Verilog type-string for `Parameter` `param`. """
        ptype = pparam.WhichOneof("value")
        if ptype == "integer":
            return "longint"
        if ptype == "double":
            return "real"
        if ptype == "string":
            return "string"
        raise ValueError

    @classmethod
    def format_param_decl(cls, name: str, param: vlsir.Param) -> str:
        """ Format a parameter-declaration """
        rv = f"parameter {name}"
        # FIXME: whether to include datatype
        # dtype = cls.format_param_type(param)
        default = cls.get_param_default(param)
        if default is not None:
            rv += f" = {default}"
        return rv

    def format_concat(self, pconc: vlsir.circuit.Concat) -> str:
        """ Format the Concatenation of several other Connections """
        # Verilog { a, b, c } concatenation format
        parts = [self.format_connection(part) for part in pconc.parts]
        return "{" + ", ".join(parts) + "}"

    @classmethod
    def format_port_decl(cls, pport: vlsir.circuit.Port) -> str:
        """ Format a `Port` declaration """

        # First retrieve and check the validity of its direction
        port_type_to_str = {
            vlsir.circuit.Port.Direction.Value("INPUT"): "input",
            vlsir.circuit.Port.Direction.Value("OUTPUT"): "output",
            vlsir.circuit.Port.Direction.Value("INOUT"): "inout",
            vlsir.circuit.Port.Direction.Value("NONE"): "NO_DIRECTION",
        }
        dir_ = port_type_to_str.get(pport.direction, None)
        if dir_ is None:
            msg = f"Invalid Verilog netlisting for unknown Port direction {pport.direction}"
            raise RuntimeError(msg)
        if dir_ == "NO_DIRECTION":
            msg = f"Invalid Verilog netlisting for undirected Port {pport}"
            raise RuntimeError(msg)

        return dir_ + " " + cls.format_signal_decl(pport.signal)

    @classmethod
    def format_signal_decl(cls, psig: vlsir.circuit.Signal) -> str:
        """ Format a `Signal` declaration """
        rv = "wire"
        if psig.width > 1:
            rv += f" [{psig.width-1}:0]"
        rv += f" {psig.name}"
        return rv

    @classmethod
    def format_port_ref(cls, pport: vlsir.circuit.Port) -> str:
        """ Format a reference to a `Port`. 
        Unlike declarations, this just requires the name of its `Signal`. """
        return cls.format_signal_ref(pport.signal)

    @classmethod
    def format_signal_ref(cls, psig: vlsir.circuit.Signal) -> str:
        """ Format a reference to a `Signal`. 
        Unlike declarations, this just requires its name. """
        return psig.name

    @classmethod
    def format_signal_slice(cls, pslice: vlsir.circuit.Slice) -> str:
        """ Format Signal-Slice `pslice` """
        if pslice.top == pslice.bot:  # Single-bit slice
            return f"{pslice.signal}[{pslice.top}]"
        return f"{pslice.signal}[{pslice.top}:{pslice.bot}]"  # Multi-bit slice

    def write_comment(self, comment: str) -> None:
        """ Verilog uses C-style line comments, beginning with `//` """
        self.write(f"// {comment}\n")



"""
# Netlister Base Class 

"""

# Std-Lib Imports
from typing import Optional, Union, IO, Dict, Iterable
from enum import Enum
from dataclasses import dataclass, field

# Local Imports
import vlsir


# Internal type shorthand
ModuleLike = Union[vlsir.circuit.Module, vlsir.circuit.ExternalModule]


class SpicePrefix(Enum):
    """ Enumerated Spice Primitives and their Instance-Name Prefixes """

    # Sub-circits, either from `Module`s or `ExternalModule`s
    SUBCKT = "x"
    # Ideal Passives
    RESISTOR = "r"
    CAPACITOR = "c"
    INDUCTOR = "l"
    # Semiconductor Devices
    MOS = "m"
    DIODE = "d"
    BIPOLAR = "q"
    # Independent Sources
    VSOURCE = "v"
    ISOURCE = "i"
    # Dependent Sources
    VCVS = "e"
    VCCS = "g"
    CCCS = "f"
    CCVS = "h"
    # Transmission Lines
    TLINE = "o"


@dataclass
class ResolvedModule:
    """ Resolved reference to a `Module` or `ExternalModule`. 
    Includes its spice-language prefix, and if user-defined its netlist-sanitized module-name. """

    module: ModuleLike
    module_name: str
    spice_prefix: SpicePrefix


@dataclass
class ResolvedParams:
    """ Resolved Instance-Parameter Values 
    Factoring in defaults, and converted to strings. 
    Largely a wrapper for `Dict[str, str]`, with accessors `get` and `pop` that raise `RuntimeError` if a key is missing. """

    inner: Dict[str, str]

    def set(self, key: str, val: str) -> None:
        """ Set the value of `key` to `val` in the resolved parameters. """
        self.inner[key] = val

    def get(self, key: str) -> str:
        """ Get the value of `key` from the resolved parameters. 
        Raises `RuntimeError` if `key` is not present. """
        if key not in self.inner:
            raise RuntimeError(f"Missing parameter {key}")
        return self.inner[key]

    def pop(self, key: str) -> str:
        """ Get the value of `key` from the resolved parameters, and remove it from the `ResolvedParams`. 
        Raises `RuntimeError` if `key` is not present. """
        if key not in self.inner:
            raise RuntimeError(f"Missing parameter {key}")
        return self.inner.pop(key)

    def pop_many(self, keys: Iterable[str]) -> Dict[str, str]:
        """ Get the values of `keys` from the resolved parameters, and remove them from the `ResolvedParams`. 
        Raises `RuntimeError` if any `key` is not present. """
        return {key: self.pop(key) for key in keys}

    def rename(self, old: str, new: str) -> None:
        """ Rename a key from `old` to `new`. 
        Raises an exception if `old` is not a key. """
        val = self.pop(old)
        self.set(new, val)

    @property
    def items(self):
        return self.inner.items

    def __bool__(self) -> bool:
        """ Boolean conversions, generally through the `not` keyword or `bool` constructor, 
        are forwarded down to the internal `inner` dictionary. """
        return bool(self.inner)


@dataclass
class Indent:
    """ 
    # Indentation Helper 
    
    Supports in-place addition and subtraction of indentation levels, e.g. via 
    ```python
    indent = Indent()
    indent += 1 # Adds one "tab", or indentation level
    indent += 1 # Adds another
    indent -= 1 # Drops back by one
    ```
    The current indentation-string is available via the `state` attribute. 
    Writers using such an indenter will likely be of the form:
    ```python
    dest.write(f"{indent.state}{content}")
    ```
    """

    # Per-"tab" indentation characters. Defaults to two spaces.
    chars: str = 2 * " "
    # Current (integer) indentation-level, in number of "tabs"
    num: int = field(init=False, default=0)
    # Current indentation-string. Always equals `num * chars`.
    state: str = field(init=False, default="")

    def __post_init_post_parse__(self) -> None:
        self.state = self.chars * self.num

    def __iadd__(self, other: int) -> None:
        """ In-place add, i.e. `indent += 1` """
        self.num += other
        self.state = self.chars * self.num
        return self

    def __isub__(self, other: int) -> None:
        """ In-place subtract, i.e. `indent -= 1` """
        self.num = self.num - other
        if self.num < 0:
            raise ValueError("Negative indentation")
        self.state = self.chars * self.num
        return self


class Netlister:
    """ # Abstract Base `Netlister` Class 

    `Netlister` is not directly instantiable, and none of its sub-classes are intended 
    for usage outside the `netlist` package. The primary API method `netlist` is designed to 
    create, use, and drop a `Netlister` instance. 
    Once instantiated a `Netlister`'s primary API method is `netlist`. 
    This writes all content in its `pkg` field to destination `dest`. 
    
    Internal methods come in two primary flavors:
    * `write_*` methods, which write to `self.dest`. These methods are generally format-specific. 
    * `format_*` methods, which return format-specific strings, but *do not* write to `dest`. 
    * `get_*` methods, which retrieve some internal data, e.g. extracting the type of a `Connection`. 
    """

    def __init__(self, pkg: vlsir.circuit.Package, dest: IO):
        self.pkg = pkg
        self.dest = dest
        self.indent = Indent(chars="  ")
        self.module_names = set()  # Netlisted Module names
        self.pmodules = dict()  # Visited proto-Modules
        self.ext_modules = dict()  # Visited ExternalModules
        self.ext_module_names = (
            dict()
        )  # Visited ExternalModule names, checked for duplicates

    def netlist(self) -> None:
        """ Primary API Method.
        Convert everything in `self.pkg` and write to `self.dest`. """

        # First visit any externally-defined Modules,
        # Ensuring we have their port-orders.
        for emod in self.pkg.ext_modules:
            self.get_external_module(emod)

        # Add some header commentary
        self.write_header()

        # Now do the real stuff,
        # Creating netlist entries for each package-defined Module
        for mod in self.pkg.modules:
            self.write_module_definition(mod)
        # And ensure all output makes it to `self.dest`
        self.dest.flush()

    def write(self, s: str) -> None:
        """ Helper/wrapper, passing to `self.dest` """
        self.dest.write(s)

    def writeln(self, s: str) -> None:
        """ Write `s` as a line, at our current `indent` level. """
        self.write(f"{self.indent.state}{s}\n")

    def get_external_module(self, emod: vlsir.circuit.ExternalModule) -> None:
        """ Visit an ExternalModule definition.
        "Netlisting" these doesn't actually write anything,
        but just stores a reference  in internal dictionary `ext_modules`
        for future references to them. """
        key = (emod.name.domain, emod.name.name)
        if key in self.ext_modules:
            raise RuntimeError(f"Invalid doubly-defined external module {emod}")
        self.ext_modules[key] = emod

    @classmethod
    def get_param_default(cls, pparam: vlsir.Param) -> Optional[str]:
        """ Get the default value of `pparam`. Returns `None` for no default. """
        if pparam.value.WhichOneof("value") is None:
            return None
        return cls.get_param_value(pparam.value)

    @classmethod
    def get_param_value(cls, ppval: vlsir.ParamValue) -> str:
        """ Get a string representation of a parameter-value """
        ptype = ppval.WhichOneof("value")
        if ptype == "integer":
            return str(int(ppval.integer))
        if ptype == "double":
            return str(float(ppval.double))
        if ptype == "string":
            return str(ppval.string)
        if ptype == "literal":
            return str(ppval.literal)
        if ptype == "prefixed":
            return cls.format_prefixed(ppval.prefixed)
        raise ValueError(f"Invalid Param type {ptype}")

    @classmethod
    def format_prefixed(cls, pre: vlsir.Prefixed) -> str:
        prefix = cls.format_prefix(pre.prefix)
        numtp = pre.WhichOneof("number")
        if numtp == "integer":
            num = str(int(pre.integer))
        elif numtp == "double":
            num = str(float(pre.double))
        elif numtp == "string":
            num = str(pre.string)
        else:
            raise ValueError(f"Invalid `Prefixed` number type {numtp}")

        # Note the "prefix" is in fact at the *end*,
        # e.g. "5u", "11K".
        # (Calling it a *pre*-fix refers to *units*, not to numeric values)
        return f"{num}{prefix}"

    @classmethod
    def format_prefix(cls, pre: vlsir.SIPrefix) -> str:
        """ Format a `SIPrefix` to a string """
        # Some of these are valid in principle, but wedunno whether netlist languages support them.
        # Eventually convert these to other prefixes, e.g. EXA => 1000 * PETA
        unsupported = {
            vlsir.SIPrefix.YOCTO,
            vlsir.SIPrefix.ZEPTO,
            vlsir.SIPrefix.ZETTA,
            vlsir.SIPrefix.YOTTA,
            vlsir.SIPrefix.CENTI,
            vlsir.SIPrefix.DECI,
            vlsir.SIPrefix.DECA,
            vlsir.SIPrefix.HECTO,
            vlsir.SIPrefix.EXA,
        }
        if pre in unsupported:
            raise RuntimeError(f"Unsupported SIPrefix for netlisting: {pre}")

        map = {
            vlsir.SIPrefix.ATTO: "a",
            vlsir.SIPrefix.FEMTO: "f",
            vlsir.SIPrefix.PICO: "p",
            vlsir.SIPrefix.NANO: "n",
            vlsir.SIPrefix.MICRO: "u",
            vlsir.SIPrefix.MILLI: "m",
            vlsir.SIPrefix.KILO: "K",
            vlsir.SIPrefix.MEGA: "M",
            vlsir.SIPrefix.GIGA: "G",
            vlsir.SIPrefix.TERA: "T",
            vlsir.SIPrefix.PETA: "P",
        }
        if pre not in map:
            raise ValueError(f"Invalid or Unsupported SIPrefix {pre}")

        return map[pre]

    @classmethod
    def get_instance_params(
        cls, pinst: vlsir.circuit.Instance, pmodule: ModuleLike
    ) -> ResolvedParams:
        """ Resolve the parameters of `pinst` to their values, including default values provided by `pmodule`. 
        Raises a `RuntimeError` if any required parameter is not defined. 

        Note this method *does not* raise errors for parameters *not specified* in `pmodule`, 
        allowing for "pass-through" parameters not explicitly defined. """

        values = dict()

        # Step through each of `pmodule`'s declared parameters first, applying defaults if necessary
        for mparam in pmodule.parameters:
            if mparam.name in pinst.parameters:  # Specified by the Instance
                inst_pval = pinst.parameters.pop(mparam.name)
                values[mparam.name] = cls.get_param_value(inst_pval)
            else:  # Not specified by the instance. Apply the default, or fail.
                pdefault = cls.get_param_default(mparam)
                if pdefault is None:
                    msg = f"Required parameter `{mparam.name}` not specified for Instance `{pinst}`"
                    raise RuntimeError(msg)
                values[mparam.name] = pdefault

        # Convert the remaining instance-provided parameters to strings
        for (pname, pval) in pinst.parameters.items():
            values[pname] = cls.get_param_value(pval)

        # And wrap the resolved values in a `ResolvedParams` object
        return ResolvedParams(values)

    @classmethod
    def get_module_name(cls, module: vlsir.circuit.Module) -> str:
        """ Create a netlist-compatible name for proto-Module `module` """

        # Create the module name
        # Replace all format-invalid characters with underscores
        name = module.name.split(".")[-1]
        for ch in name:
            if not (ch.isalpha() or ch.isdigit() or ch == "_"):
                name = name.replace(ch, "_")
        return name

    def resolve_reference(self, ref: vlsir.utils.Reference) -> ResolvedModule:
        """ Resolve the `ModuleLike` referent of `ref`. """

        if ref.WhichOneof("to") == "local":  # Internally-defined Module
            module = self.pmodules.get(ref.local, None)
            if module is None:
                raise RuntimeError(f"Invalid undefined Module {ref.local} ")
            return ResolvedModule(
                module=module,
                module_name=self.get_module_name(module),
                spice_prefix=SpicePrefix.SUBCKT,
            )

        if ref.WhichOneof("to") == "external":  # Defined outside package

            # First check the priviledged/ internally-defined domains
            if ref.external.domain == "vlsir.primitives":
                # Built-in primitive. Load its definition from the `vlsir.primitives` (python) module.
                name = ref.external.name
                module = vlsir.primitives.dct.get(ref.external.name, None)
                if module is None:
                    raise RuntimeError(f"Invalid undefined primitive {ref.external}")

                # Mapping from primitive-name to spice-prefix
                prefixes = dict(
                    resistor=SpicePrefix.RESISTOR,
                    capacitor=SpicePrefix.CAPACITOR,
                    inductor=SpicePrefix.INDUCTOR,
                    vdc=SpicePrefix.VSOURCE,
                    vpulse=SpicePrefix.VSOURCE,
                    vpwl=SpicePrefix.VSOURCE,
                    vsin=SpicePrefix.VSOURCE,
                    isource=SpicePrefix.ISOURCE,
                    vcvs=SpicePrefix.VCVS,
                    vccs=SpicePrefix.VCCS,
                    cccs=SpicePrefix.CCCS,
                    ccvs=SpicePrefix.CCVS,
                    mos=SpicePrefix.MOS,
                    bipolar=SpicePrefix.BIPOLAR,
                    diode=SpicePrefix.DIODE,
                )

                if name not in prefixes:
                    raise ValueError(f"Unsupported or Invalid Ideal Primitive {ref}")

                return ResolvedModule(
                    module=module,
                    module_name=module.name.name,
                    spice_prefix=prefixes[name],
                )

            if ref.external.domain == "hdl21.primitives":
                msg = f"Invalid direct-netlisting of physical `hdl21.Primitive` `{ref.external.name}`. "
                msg += "Either compile to a target technology, or replace with an `ExternalModule`. "
                raise RuntimeError(msg)

            else:  # Externally-Defined, External-Domain `ExternalModule`
                key = (ref.external.domain, ref.external.name)
                module = self.ext_modules.get(key, None)
                if module is None:
                    msg = f"Invalid Instance of undefined External Module {key}"
                    raise RuntimeError(msg)
                # Check for duplicate names which would conflict from other namespaces
                module_name = ref.external.name
                if (
                    module_name in self.ext_module_names
                    and self.ext_module_names[module_name] is not module
                ):
                    msg = f"Conflicting ExternalModule definitions {module} and {self.ext_module_names[module_name]}"
                    raise RuntimeError(msg)
                self.ext_module_names[module_name] = module
                return ResolvedModule(
                    module=module,
                    module_name=module_name,
                    spice_prefix=SpicePrefix.SUBCKT,
                )

        # Not a Module, not an ExternalModule, not sure what it is
        raise ValueError(f"Invalid Module reference {ref}")

    def format_connection(self, pconn: vlsir.circuit.Connection) -> str:
        """ Format a `Connection` reference. 
        Does not *declare* any new connection objects, but generates references to existing ones. """
        # Connections are a proto `oneof` union
        # which includes signals, slices, and concatenations.
        # Figure out which to import

        stype = pconn.WhichOneof("stype")
        if stype == "sig":
            return self.format_signal_ref(pconn.sig)
        if stype == "slice":
            return self.format_signal_slice(pconn.slice)
        if stype == "concat":
            return self.format_concat(pconn.concat)
        raise ValueError(f"Invalid Connection Type {stype} for Connection {pconn}")

    def write_header(self) -> None:
        """ Write header commentary 
        This proves particularly important for many Spice-like formats, 
        which *always* interpret the first line of an input-file as a comment (among many other dumb things). 
        So, always write one there right off the bat. """

        if self.pkg.domain:
            self.write_comment(f"circuit.Package {self.pkg.domain}")
        else:
            self.write_comment(f"Anonymous circuit.Package")
        self.write_comment(f"Written by {self.__class__.__name__}")
        self.write_comment("")
        self.writeln("")

    """ 
    Virtual `format` Methods 
    """

    @classmethod
    def format_param_decl(cls, name: str, param: vlsir.Param) -> str:
        """ Format a named `Parameter` definition """
        raise NotImplementedError

    @classmethod
    def format_port_decl(cls, pport: vlsir.circuit.Port) -> str:
        """ Format a declaration of a `Port` """
        raise NotImplementedError

    @classmethod
    def format_port_ref(cls, pport: vlsir.circuit.Port) -> str:
        """ Format a reference to a `Port` """
        raise NotImplementedError

    @classmethod
    def format_signal_decl(cls, psig: vlsir.circuit.Signal) -> str:
        """ Format a declaration of Signal `psig` """
        raise NotImplementedError

    @classmethod
    def format_signal_ref(cls, psig: vlsir.circuit.Signal) -> str:
        """ Format a reference to Signal `psig` """
        raise NotImplementedError

    @classmethod
    def format_signal_slice(cls, pslice: vlsir.circuit.Slice) -> str:
        """ Format Signal-Slice `pslice` """
        raise NotImplementedError

    def format_concat(self, pconc: vlsir.circuit.Concat) -> str:
        """ Format the Concatenation of several other Connections """
        raise NotImplementedError

    @classmethod
    def format_bus_bit(cls, index: Union[int, str]) -> str:
        """ Format bus-bit `index` """
        raise NotImplementedError

    """ 
    Virtual `write` Methods 
    """

    def write_comment(self, comment: str) -> None:
        """ Format and string a string comment. 
        "Line comments" are the sole supported variety, which fit within a line, and extend to the end of that line. 
        The `write_comment` method assumes responsibility for closing the line. """
        raise NotImplementedError

    def write_module_definition(self, pmodule: vlsir.circuit.Module) -> None:
        """ Write Module `module` """
        raise NotImplementedError

    def write_instance(self, pinst: vlsir.circuit.Instance) -> str:
        """ Write Instance `pinst` """
        raise NotImplementedError

    """ 
    Other Virtual Methods 
    """

    @property
    def enum(self):
        """ Get our entry in the `NetlistFormat` enumeration """
        raise NotImplementedError




""" 
# Spice-Class Simulator Interface 

Base class(es), utilities, and shared functionality for simulators. 
"""

# Std-Lib Imports
import os, tempfile, subprocess
from typing import Union, Optional
from enum import Enum
from pathlib import Path
from textwrap import dedent
from dataclasses import dataclass, field

# Local/ Project Dependencies
import vlsir.spice_pb2 as vsp
from vlsir.spice_pb2 import *  # Not used here intentionally, but "re-exported"
from . import sim_data as sd


class ResultFormat(Enum):
    """ Enumerated Result Formats """

    SIM_DATA = "sim_data"  # Types defined in the `sim_data` module
    VLSIR_PROTO = "vlsir_proto"  # `vsp.SimResults` and related protobuf-defined types


# Union of the two result-types, returned by many simulation methods
SimResultUnion = Union[vsp.SimResult, sd.SimResult]


class AnalysisType(Enum):
    """ Enumerated Analysis-Types 
    Corresponding to the entries in the `Analysis` type-union. """

    DC = "dc"
    AC = "ac"
    TRAN = "tran"
    MONTE = "monte"
    SWEEP = "sweep"
    CUSTOM = "custom"


class SupportedSimulators(Enum):
    """ Enumerated, Internally-Defined Spice-Class Simulators """

    SPECTRE = "spectre"
    XYCE = "xyce"


def default() -> Optional[SupportedSimulators]:
    """ Get the default simulator, for this Python-process and its environment. 
    This largely consists of a priority-ordered walk through `SupportedSimulators`, 
    returning the first whose `available()` method indicates its availability. 
    Returns `None` if no such simulator appears available. """

    from .spectre import available as spectre_available
    from .xyce import available as xyce_available

    if spectre_available():
        return SupportedSimulators.SPECTRE
    if xyce_available():
        return SupportedSimulators.XYCE
    return None  # Nothing found


@dataclass
class SimOptions:
    """ Options to `vsp.Sim` which are *not* passed along to the simulator. 
    I.e. options which effect the Python-process invoking simulation, 
    but not the internals of the simulation itself. """

    # Simulator. FIXME: debatable whether to include this.
    simulator: SupportedSimulators = field(default_factory=default)

    # In-memory results format
    fmt: ResultFormat = ResultFormat.VLSIR_PROTO

    # Simulation run-directory. Uses a `tempdir` if unspecified.
    rundir: Optional[os.PathLike] = None


def sim(inp: vsp.SimInput, opts: Optional[SimOptions] = None) -> SimResultUnion:
    """
    Execute a `vlir.spice.Sim`. 
    Dispatches across `SupportedSimulators` specified in `SimOptions` `opts`. 
    Uses the default `Simulator` as detected by the `default` method if no `simulator` is specified.
    """
    if opts is None:  # Create the default `SimOptions`
        opts = SimOptions()

    if opts.simulator is None:  # If we didn't specify or find a simulator, fail.
        msg = f"vlsirtools.spice: No Supported Simulators available for call to `sim()`"
        raise RuntimeError(msg)

    # Get the per-simulator callable
    if opts.simulator == SupportedSimulators.XYCE:
        from .xyce import sim
    elif opts.simulator == SupportedSimulators.SPECTRE:
        from .spectre import sim
    else:
        raise ValueError(f"Unsupported simulator: {opts.simulator}")

    # And do the real work, invoking the target simulator
    return sim(inp, opts)


class Sim:
    """
    # Simulator State Base-Class 

    Shared basic functionality for simulators, including: 
    * Creating run-directories, and named files within them 
    * Finding a valid top-level module 
    * Exception/ Error types for common simulator errors

    Each `SupportedSimulator` is generally implemented as a (python) module including:

    * A `sim` free-standing function, and 
    * An implementation sub-class of `Sim` which manages run-state and simulator-specific behavior 

    The latter generally should (and generally does) call the classmethod `Sim.sim`, 
    which creates and navigates to a specified or temporary directory, 
    and cleans up all simulation files after itself. 

    The interface between the `Sim` base-class and its implementing sub-class is such that:

    * `run` is the primary entry point, implemented by this base class. 
      * After initial checks and setup, `run` hands off to a sub-class-specific `_run` method.
    * When used as a context manager via `with`, all internal activity occurs within the simulation's temporary directory.
      * System-calls to create and manage files can be made without knowledge of this directory.
      * E.g. `open('netlist', 'w')` will land `netlist` in the temporary directory.
    * All other methods are implemented by the sub-classes. 
      * No requirements or conventions are placed on their names or activities. 
      * Such methods will generally do things including: writing simulator-specific netlist-content, launching a sim process, parsing results. 
      * At no point should the sub-classes need to know any more about the `Sim` base-class, or call any of its `super` methods. 
    """

    @classmethod
    def enum(cls) -> SupportedSimulators:
        raise NotImplementedError

    @classmethod
    def sim(
        cls, inp: vsp.SimInput, opts: Optional[SimOptions] = None
    ) -> SimResultUnion:
        """ Sim-invoking class method. 
        Creates an instance of `cls` as a context manager, run in its simulation directory. 
        This should be invoked by typical implementations of a free-standing `sim` function. """

        if opts is None:  # Create the default `SimOptions`
            opts = SimOptions(simulator=cls.enum())

        # Create the simulation class, and execute its main `run` method in `rundir`
        with cls(inp=inp, rundir=opts.rundir) as sim:
            results = sim.run()

        # FIXME: And handle output formatting, when these `Sim` classes are actually compatible
        # if opts.fmt == ResultFormat.VLSIR_PROTO:
        #     return results.to_proto()
        return results

    def __init__(self, inp: vsp.SimInput, rundir: Optional[os.PathLike] = None) -> None:
        self.inp = inp
        self.rundir = rundir
        self.prevdir = os.getcwd()
        self.top = None

    def __enter__(self) -> "Sim":
        """ On entry, create and navigate to a directory for the sim's files.  """
        if self.rundir is not None:
            self.rundir = Path(self.rundir).absolute()
            if not self.rundir.exists():
                os.makedirs(self.rundir)
            os.chdir(self.rundir)
        else:  # Create a new temp directory
            self.rundir = tempfile.TemporaryDirectory()
            os.chdir(self.rundir.name)
        return self

    def __exit__(self, _type, _value, _traceback):
        """ On exit, clean up our temporary directory, and navigate back to its predecessor.  """
        os.chdir(self.prevdir)
        if isinstance(self.rundir, tempfile.TemporaryDirectory):
            self.rundir.cleanup()

    def validate_top(self) -> None:
        """ Ensure that the `top` module exists,
        and adheres to the "Spice top-level" port-interface: a single port for ground / VSS / node-zero. 
        Sets the `self.top` attribute when successful, or raises a `RuntimeError` when not. """

        if not self.inp.top:
            raise RuntimeError(f"No top-level module specified")

        found = False
        for module in self.inp.pkg.modules:
            if module.name == self.inp.top:
                found = True
                self.top = module
                if len(module.ports) != 1:
                    msg = f"`vlsir.SimInput` top-level module {self.inp.top} must have *one* (VSS) port - has {len(module.ports)} ports [{module.ports}]"
                    raise RuntimeError(msg)
                break

        if not found:
            names = [m.name for m in self.inp.pkg.modules]
            raise RuntimeError(
                f"Top-level module `{self.inp.top}` not found among Modules {names}"
            )

    def run(self) -> SimResultUnion:
        """ Primary invocation method. 
        Run the specified `SimInput` in directory `self.rundir`, returning its results. 
        Performs initial setup, then hands off to the simulator-specific sub-class's `_run` method. """

        # Setup
        self.validate_top()

        # And hand off to the simulator-specific sub-class's `_run` method.
        return self._run()

    def _run(self) -> SimResultUnion:
        raise NotImplementedError("`Sim` subclasses must implement `_run`")


class SimError(Exception):
    """ Exception raised when a simulation fails. """

    pass


class SimProcessError(SimError):
    """ Exception raised when an external simulator process fails. """

    def __init__(self, sim: Sim, e: subprocess.CalledProcessError) -> None:
        s = dedent(
            f"""
            rundir: {str(sim.rundir)}
            stdout: {str(e.stdout)}
            stderr: {str(e.stderr)}
        """
        )
        super().__init__(s)
        self.rundir = str(sim.rundir)
        self.stdout = e.stdout
        self.stderr = e.stderr




"""
A mirror of Spice Proto's Analysis Result objects,
but using python data classes and numpy arrays.

Also provides round-tripping utilities between the two
TODO: Go from proto -> sim_result
"""


from dataclasses import dataclass
import numpy as np
from typing import List, Mapping, Union


import vlsir


@dataclass
class OpResult:
    analysis_name: str
    data: Mapping[str, float]
    vlsir_type: str = "op"

    def to_proto(self) -> vlsir.spice.OpResult:
        res = vlsir.spice.OpResult(analysis_name=self.analysis_name)
        for k, v in self.data.items():
            res.signals.append(k)
            res.data.append(v)
        return res


@dataclass
class DcResult:
    analysis_name: str
    indep_name: str
    data: Mapping[str, np.ndarray]
    measurements: Mapping[str, float]
    vlsir_type: str = "dc"

    def to_proto(self) -> vlsir.spice.DcResult:
        res = vlsir.spice.DcResult(
            analysis_name=self.analysis_name, indep_name=self.indep_name
        )
        for k, v in self.data.items():
            res.signals.append(k)
            for d in v:
                res.data.append(d)
        # TODO Add support for measurements
        return res


@dataclass
class TranResult:
    analysis_name: str
    data: Mapping[str, np.ndarray]
    measurements: Mapping[str, float]
    vlsir_type: str = "tran"

    def to_proto(self) -> vlsir.spice.TranResult:
        res = vlsir.spice.TranResult(analysis_name=self.analysis_name)
        for k, v in self.data.items():
            res.signals.append(k)
            for d in v:
                res.data.append(d)
        # TODO Add support for measurements
        return res


@dataclass
class AcResult:
    analysis_name: str
    freq: np.ndarray
    data: Mapping[str, np.ndarray]
    measurements: Mapping[str, float]
    vlsir_type: str = "ac"

    def to_proto(self) -> vlsir.spice.AcResult:
        raise NotImplementedError


# Type alias for the union of each result-type
AnalysisResult = Union[AcResult, DcResult, TranResult, OpResult]


@dataclass
class SimResult:
    """ Results from a Mult-Analysis `Sim` """

    an: List[AnalysisResult]

    def to_proto(self) -> vlsir.spice.SimResult:
        res = vlsir.spice.SimResult()
        for a in self.an:
            res.an.append(vlsir.spice.AnalysisResult(**{a.vlsir_type: a.to_proto()}))
        return res

    def __getitem__(self, key: int) -> AnalysisResult:
        return self.an[key]




"""
Xyce Implementation of `vsp.Sim`
"""

# Std-Lib Imports
import subprocess, random, shutil, csv
from typing import List, Tuple, IO, Optional, Dict
from glob import glob

# Local/ Project Dependencies
import vlsir.spice_pb2 as vsp
from ..netlist import netlist, XyceNetlister
from .spice import (
    Sim,
    SimProcessError,
    ResultFormat,
    SimOptions,
    SupportedSimulators,
    SimResultUnion,
)


# Module-level configuration. Over-writeable by sufficiently motivated users.
XYCE_EXECUTABLE = "Xyce"  # The simulator executable invoked. If over-ridden, likely for sake of a specific path or version.


def available() -> bool:
    """ Boolean indication of whether the current running environment includes the simulator executable on its path. """
    return shutil.which(XYCE_EXECUTABLE) is not None


def sim(inp: vsp.SimInput, opts: Optional[SimOptions] = None) -> SimResultUnion:
    """ # Primary Simulation Method """

    if opts is None:  # Create the default options
        opts = SimOptions(simulator=SupportedSimulators.XYCE)

    if opts.fmt != ResultFormat.VLSIR_PROTO:
        raise RuntimeError(f"Unsupported ResultFormat: {opts.fmt} for Xyce")

    return XyceSim.sim(inp, opts)


class XyceSim(Sim):
    """ 
    State and execution logic for a Xyce-call to `vsp.Sim`. 
    
    Xyce can, in principle, run multiple analyses per process, 
    but seems to commonly confuse outputs or disallow saving them
    from multiple analyses. 

    Execution therefore instead occurs one Xyce-process per analysis. 
    Results from each analysis-process are collated into a single `SimResult`. 
    """

    @classmethod
    def enum(cls) -> SupportedSimulators:
        return SupportedSimulators.XYCE

    def _run(self) -> vsp.SimResult:
        """ Run the specified `SimInput` in directory `self.rundir`, 
        returning its results. """

        netlist_file = open("dut", "w")
        netlist(pkg=self.inp.pkg, dest=netlist_file, fmt="xyce")

        # Write the top-level instance
        netlist_file.write(
            f"xtop 0 {XyceNetlister.get_module_name(self.top)} ; Top-Level DUT \n\n"
        )

        if self.inp.opts:
            raise NotImplementedError(f"SimInput Options")

        # Write each control element
        self.write_control_elements(netlist_file)

        # Flush the netlist to disk before handing off to analyses
        netlist_file.flush()

        # Run each analysis in the input
        results = vsp.SimResult()
        for an in self.inp.an:
            results.an.append(self.analysis(an))
        return results

    def write_control_elements(self, netlist_file: IO) -> None:
        for ctrl in self.inp.ctrls:
            inner = ctrl.WhichOneof("ctrl")
            if inner == "include":
                netlist_file.write(f".include '{ctrl.include.path}' \n")
            elif inner == "lib":
                netlist_file.write(f".lib {ctrl.lib.path} {ctrl.lib.section} \n")
            elif inner == "param":
                line = f".param {ctrl.param.name}={XyceNetlister.get_param_value(ctrl.param.value)} \n"
                netlist_file.write(line)
            elif inner == "meas":
                line = f".meas {ctrl.meas.analysis_type} {ctrl.meas.name} {ctrl.meas.expr} \n"
                netlist_file.write(line)
            elif inner == "literal":
                netlist_file.write(ctrl.literal + "\n")
            elif inner in ("save"):
                raise NotImplementedError(
                    f"Unimplemented control card {ctrl} for {self}"
                )  # FIXME!
            else:
                raise RuntimeError(f"Unknown control type: {inner}")

    def analysis(self, an: vsp.Analysis) -> vsp.AnalysisResult:
        """ Execute a `vsp.Analysis`, returning its `vsp.AnalysisResult`. """

        # `Analysis` is a Union (protobuf `oneof`) of the analysis-types.
        # Unwrap it, and dispatch based on the type.
        AR = vsp.AnalysisResult  # Quick shorthand
        inner = an.WhichOneof("an")

        if inner == "op":
            return AR(op=self.op(an.op))
        if inner == "dc":
            return AR(dc=self.dc(an.dc))
        if inner == "ac":
            return AR(ac=self.ac(an.ac))
        if inner == "tran":
            return AR(tran=self.tran(an.tran))
        if inner in ("sweep", "monte", "custom"):
            raise NotImplementedError(f"{inner} not implemented")
        raise RuntimeError(f"Unknown analysis type: {inner}")

    def ac(self, an: vsp.AcInput) -> vsp.AcResult:
        """ Run an AC analysis. """

        # Unpack the `AcInput`
        analysis_name = an.analysis_name or "ac"
        if len(an.ctrls):
            raise NotImplementedError  # FIXME!

        # Copy and append to the existing DUT netlist
        shutil.copy("dut", f"{analysis_name}.sp")
        netlist = open(f"{analysis_name}.sp", "a")

        # Write the analysis command
        npts = an.npts
        fstart = an.fstart
        fstop = an.fstop
        netlist.write(f".ac DEC {npts} {fstart} {fstop} \n\n")

        # FIXME: always saving everything, no matter what
        # Note `csv` output-formatting is encoded here
        netlist.write(".print ac format=csv v(*) i(*) \n\n")

        # And don't forget - the thing SPICE can't live without - END!
        netlist.write(".end \n\n")
        netlist.flush()

        # Do the real work, running the simulation
        self.run_xyce_process(analysis_name)

        # Read the results from CSV
        with open(f"{analysis_name}.sp.FD.csv", "r") as csv_handle:
            (signals, data) = read_csv(csv_handle)

        # Separate Frequency vector
        n_sigs = len(signals)  # Get length of signals vector because...
        freq = data[::n_sigs]  # ...every n-th data pt is a freq pt
        del data[::n_sigs]  # Clean the list of data of all frequencies
        signals.pop(0)  # Remove "Frequency" from list of signals

        # Build ComplexNumbers for each data point
        reals = data[::2]
        imags = data[1::2]
        if not len(reals) == len(imags):  # Sanity check
            raise RuntimeError("Unpaired complex number in data")
        cplx_data = [
            vsp.ComplexNum(re=real, im=imag) for real, imag in zip(reals, imags)
        ]

        # Parse any scalar measurement results
        measurements = parse_measurements(analysis_name)

        # And arrange them in an `AcResult`
        return vsp.AcResult(
            freq=freq, signals=signals, data=cplx_data, measurements=measurements
        )

    def dc(self, an: vsp.DcInput) -> vsp.DcResult:
        """ Run a DC analysis. """

        # Unpack the `DcInput`
        analysis_name = an.analysis_name or "dc"

        if len(an.ctrls):
            raise NotImplementedError  # FIXME!

        # Copy and append to the existing DUT netlist
        shutil.copy("dut", f"{analysis_name}.sp")
        netlist = open(f"{analysis_name}.sp", "a")

        # Write the analysis command
        param = an.indep_name
        ## Interpret the sweep
        sweep_type = an.sweep.WhichOneof("tp")
        if sweep_type == "linear":
            sweep = an.sweep.linear
            netlist.write(
                f".dc LIN {param} {sweep.start} {sweep.stop} {sweep.step}\n\n"
            )
        elif sweep_type == "log":
            sweep = an.sweep.log
            netlist.write(
                f".dc DEC {param} {sweep.start} {sweep.stop} {sweep.npts}\n\n"
            )
        elif sweep_type == "points":
            sweep = an.sweep.points
            netlist.write(
                f".dc {param} LIST {' '.join([str(pt) for pt in sweep.points])}\n\n"
            )
        else:
            raise ValueError("Invalid sweep type")

        # FIXME: always saving everything, no matter what
        # Note `csv` output-formatting is encoded here
        netlist.write(".print dc format=csv v(*) i(*) \n\n")

        # And don't forget - the thing SPICE can't live without - END!
        netlist.write(".end \n\n")
        netlist.flush()

        # Do the real work, running the simulation
        self.run_xyce_process(analysis_name)

        # Read the results from CSV
        with open(f"{analysis_name}.sp.csv", "r") as csv_handle:
            (signals, data) = read_csv(csv_handle)

        # Parse any scalar measurement results
        measurements = parse_measurements(analysis_name)

        # And arrange them in an `OpResult`
        return vsp.DcResult(signals=signals, data=data, measurements=measurements)

    def op(self, an: vsp.OpInput) -> vsp.OpResult:
        """ Run an operating-point analysis. 
        Xyce describes the `.op` analysis as "partially supported". 
        Here the `vsp.Op` analysis is mapped to DC, with a dummy sweep. """

        # Unpack the `OpInput`
        analysis_name = an.analysis_name or "op"
        if len(an.ctrls):
            raise NotImplementedError  # FIXME!

        # Copy and append to the existing DUT netlist
        shutil.copy("dut", f"{analysis_name}.sp")
        netlist = open(f"{analysis_name}.sp", "a")

        # Create the dummy parameter, and "sweep" a single value of it
        dummy_param = f"_dummy_{random.randint(0,65536)}_"
        netlist.write(f".param {dummy_param}=1 \n\n")

        # Write the analysis command
        netlist.write(f".dc {dummy_param} 1 1 1 \n\n")

        # FIXME: always saving everything, no matter what
        # Note `csv` output-formatting is encoded here
        netlist.write(".print dc format=csv v(*) i(*) \n\n")

        # And don't forget - the thing SPICE can't live without - END!
        netlist.write(".end \n\n")
        netlist.flush()

        # Do the real work, running the simulation
        self.run_xyce_process(analysis_name)

        # Read the results from CSV
        with open(f"{analysis_name}.sp.csv", "r") as csv_handle:
            (signals, data) = read_csv(csv_handle)

        # And arrange them in an `OpResult`
        return vsp.OpResult(signals=signals, data=data)

    def tran(self, an: vsp.TranInput) -> vsp.TranResult:
        """ Run a transient analysis. """

        # Extract fields from our `TranInput`
        analysis_name = an.analysis_name or "tran"

        # Why not make tstop/tstep required?
        if not an.tstop or not an.tstep:
            raise ValueError("tstop and tstep must be defined")
        tstop = an.tstop
        tstep = an.tstep
        if len(an.ic):
            raise NotImplementedError
        if len(an.ctrls):
            raise NotImplementedError

        # Copy and append to the existing DUT netlist
        shutil.copy("dut", f"{analysis_name}.sp")
        netlist = open(f"{analysis_name}.sp", "a")

        # Write the analysis command
        netlist.write(f".tran {tstep} {tstop} \n\n")

        # FIXME: always saving everything, no matter what
        # Note `csv` output-formatting is encoded here
        netlist.write(".print tran format=csv v(*) i(*) \n\n")

        # And don't forget - the thing SPICE can't live without - END!
        netlist.write(".end \n\n")
        netlist.flush()

        # Do the real work, running the simulation
        self.run_xyce_process(analysis_name)

        # Parse and organize our results
        # First pull them in from CSV
        with open(f"{analysis_name}.sp.csv", "r") as csv_handle:
            (signals, data) = read_csv(csv_handle)

        # Parse any scalar measurement results
        measurements = parse_measurements(analysis_name)

        # And organize them into a `TranResult` message
        return vsp.TranResult(
            analysis_name=analysis_name,
            signals=signals,
            data=data,
            measurements=measurements,
        )

    def run_xyce_process(self, name: str):
        """ Run a `Xyce` sub-process, collecting terminal output. """

        try:
            _result = subprocess.run(
                f"{XYCE_EXECUTABLE} {name}.sp ",
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                shell=True,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise SimProcessError(self, e)
        except Exception as e:
            raise


def read_csv(handle: IO) -> Tuple[List[str], List[float]]:
    """ Read a text-header + float CSV from file-handle `handle`. """

    # Get the header-list of strings
    header_line = handle.readline().strip()
    headers = header_line.split(",")

    # The remaining rows are data-values. Append them to the (single-dimension) list of results.
    data = []
    results_csv = csv.reader(handle, quoting=csv.QUOTE_NONNUMERIC)
    for row in results_csv:
        data.extend(row)

    # And return the two as a tuple
    return (headers, data)


def parse_meas(file: IO) -> Dict[str, float]:
    """ Parse an (open) measurement-file to a {name: value} dictionary. """
    rv = {}
    for line in file.readlines():
        contents = line.split()
        if len(contents) != 3 or contents[1] != "=":
            raise RuntimeError(f"Invalid line in Xyce measurements: {line}")
        name, val = contents[0], float(contents[2])
        rv[name] = val
    return rv


def parse_measurements(analysis_name: str) -> Dict[str, float]:
    # FIXME: the *input* should really be dictating whether we have measurements.
    # For now, we just search for any matching filenames via `glob`
    meas_glob = glob(f"*{analysis_name}*.m*0")
    if len(meas_glob) > 1:
        raise RuntimeError(f"Unsupported multiple measurement results for {self}")
    if len(meas_glob) == 1:
        measurements = parse_meas(open(meas_glob[0], "r"))
        return {k.lower(): v for k, v in measurements.items()}
    # No measurement-file, return an empty result
    return {}




"""
# Vlsir Python Tools
Unit Tests
"""

import pytest
import vlsirtools

from io import StringIO
from typing import Dict


def test_version():
    assert vlsirtools.__version__ == "1.0.0rc0"


def test_netlist1():
    # Test netlisting, including instantiating primitives

    from vlsir import Reference, QualifiedName, ParamValue
    from vlsir.circuit_pb2 import (
        Module,
        Signal,
        Connection,
        Port,
        Instance,
        Package,
    )

    def _prim(name: str) -> Reference:
        # Shorthand for a `Reference` to primitive `name`
        return Reference(external=QualifiedName(domain="vlsir.primitives", name=name))

    def _conns() -> Dict:
        # Shorthand for connections between node "1" and "VSS", used for many instances below.
        return dict(
            p=Connection(sig=Signal(name="vvv", width=1)),
            n=Connection(sig=Signal(name="VSS", width=1)),
        )

    pkg = Package(
        domain="vlsir.tests.test_netlist1",
        modules=[
            Module(
                name="mid",
                ports=[
                    Port(direction="NONE", signal=Signal(name="vvv", width=1)),
                    Port(direction="NONE", signal=Signal(name="VSS", width=1)),
                ],
                signals=[],
                instances=[
                    Instance(
                        name="r",
                        module=_prim("resistor"),
                        connections=_conns(),
                        parameters=dict(r=ParamValue(double=1e3)),
                    ),
                    Instance(
                        name="c",
                        module=_prim("capacitor"),
                        connections=_conns(),
                        parameters=dict(c=ParamValue(double=1e-15)),
                    ),
                    Instance(
                        name="l",
                        module=_prim("inductor"),
                        connections=_conns(),
                        parameters=dict(l=ParamValue(double=1e-9)),
                    ),
                    Instance(
                        name="v",
                        module=_prim("vdc"),
                        connections=_conns(),
                        parameters=dict(dc=ParamValue(double=1.1)),
                    ),
                    Instance(
                        name="i",
                        module=_prim("isource"),
                        connections=_conns(),
                        parameters=dict(dc=ParamValue(double=1e-6)),
                    ),
                    Instance(
                        name="m",
                        module=_prim("mos"),
                        connections=dict(
                            d=Connection(sig=Signal(name="VSS", width=1)),
                            g=Connection(sig=Signal(name="VSS", width=1)),
                            s=Connection(sig=Signal(name="VSS", width=1)),
                            b=Connection(sig=Signal(name="VSS", width=1)),
                        ),
                        parameters=dict(modelname=ParamValue(string="some_model_name")),
                    ),
                    Instance(
                        name="q",
                        module=_prim("bipolar"),
                        connections=dict(
                            c=Connection(sig=Signal(name="VSS", width=1)),
                            b=Connection(sig=Signal(name="VSS", width=1)),
                            e=Connection(sig=Signal(name="VSS", width=1)),
                        ),
                        parameters=dict(modelname=ParamValue(string="some_model_name")),
                    ),
                    Instance(
                        name="d",
                        module=_prim("diode"),
                        connections=_conns(),
                        parameters=dict(modelname=ParamValue(string="some_model_name")),
                    ),
                ],
            ),
            Module(
                name="top",
                ports=[Port(direction="NONE", signal=Signal(name="VSS", width=1)),],
                signals=[Signal(name="vvv", width=1)],
                instances=[
                    Instance(
                        name="imid",
                        module=Reference(local="mid"),
                        connections=dict(
                            vvv=Connection(sig=Signal(name="vvv", width=1)),
                            VSS=Connection(sig=Signal(name="VSS", width=1)),
                        ),
                    )
                ],
            ),
        ],
    )
    dest = StringIO()
    vlsirtools.netlist(pkg=pkg, dest=dest, fmt="spice")


@pytest.mark.xfail(reason="Hdl21 dependency pending deprecation")
def test_netlist_hdl21_ideal1():
    # Test netlisting an `hdl21.ideal` element

    from vlsir.circuit_pb2 import Module, Signal, Connection, Port, Instance, Package
    from vlsir.utils_pb2 import Reference, QualifiedName

    pkg = Package(
        domain="vlsir.tests.test_netlist_hdl21_ideal1",
        modules=[
            Module(
                name="mid",
                ports=[
                    Port(direction="NONE", signal=Signal(name="vvv", width=1)),
                    Port(direction="NONE", signal=Signal(name="VSS", width=1)),
                ],
                signals=[],
                instances=[
                    Instance(
                        name="r1",
                        module=Reference(
                            external=QualifiedName(
                                domain="hdl21.ideal",
                                name="IdealResistor",  # FIXME: being deprecated
                            )
                        ),
                        connections=dict(
                            p=Connection(sig=Signal(name="vvv", width=1)),
                            n=Connection(sig=Signal(name="VSS", width=1)),
                        ),
                    ),
                ],
            )
        ],
    )
    dest = StringIO()
    vlsirtools.netlist(pkg=pkg, dest=dest, fmt="spice")


def empty_testbench_package():
    """ Create and return a `circuit.Package` with a single, (near) empty testbench module. 
    Some simulators *really* don't like empty DUT content, and others don't like singly-connected nodes. 
    So the simplest test-bench is two resistors, in parallel, between ground and a single "other node". """

    from vlsir import Reference, QualifiedName, ParamValue
    from vlsir.circuit_pb2 import (
        Module,
        Signal,
        Connection,
        Port,
        Instance,
        Package,
    )

    def _r(name: str) -> Instance:
        # Create a canned instance of `vlsir.primitives.resistor` with instance-name `name`
        return Instance(
            name=name,  # <= Instance name argument here
            module=Reference(
                external=QualifiedName(domain="vlsir.primitives", name="resistor")
            ),
            connections=dict(
                p=Connection(sig=Signal(name="the_other_node", width=1)),
                n=Connection(sig=Signal(name="VSS", width=1)),
            ),
            parameters=dict(r=ParamValue(double=1e3)),
        )

    return Package(
        domain="vlsirtools.tests.empty_testbench_package",
        modules=[
            Module(
                name="empty_testbench",
                ports=[Port(direction="NONE", signal=Signal(name="VSS", width=1)),],
                signals=[Signal(name="the_other_node", width=1),],
                instances=[_r("r1"), _r("r2"),],
            )
        ],
    )


@pytest.mark.skipif(
    not vlsirtools.spectre.available(), reason="No spectre installation on path",
)
def test_spectre1():
    """ Test an empty-input call to the `vlsir.spice.Sim` interface to `spectre`. """
    from vlsir.spice_pb2 import SimInput, SimResult
    from vlsirtools.spectre import sim

    results = sim(SimInput(top="empty_testbench", pkg=empty_testbench_package(),))
    assert isinstance(results, SimResult)


@pytest.mark.skipif(
    not vlsirtools.xyce.available(), reason="No Xyce installation on path",
)
def test_xyce1():
    """ Test an empty-input call to the `vlsir.spice.Sim` interface to `xyce`. """
    from vlsir.spice_pb2 import SimInput, SimResult
    from vlsirtools.xyce import sim

    results = sim(SimInput(top="empty_testbench", pkg=empty_testbench_package(),))
    assert isinstance(results, SimResult)


def test_xyce_import():
    # Just test importing from the `vlsirtools.xyce` path
    from vlsirtools.xyce import sim


def test_spectre_import():
    # Just test importing from the `vlsirtools.spectre` path
    from vlsirtools.spectre import sim




"""
Spectre Implementation of `vlsir.spice.Sim`
"""

# Std-Lib Imports
import subprocess, re, shutil, glob
import numpy as np
from typing import Tuple, Any, Mapping, Optional, IO, Dict
from warnings import warn

# Local/ Project Dependencies
import vlsir.spice_pb2 as vsp
from ..netlist import netlist
from ..netlist.spectre import SpectreNetlister
from .sim_data import TranResult, OpResult, SimResult  ##, AcResult, DcResult,
from .spice import (
    Sim,
    SimProcessError,
    ResultFormat,
    SimOptions,
    SupportedSimulators,
    SimResultUnion,
)

# Module-level configuration. Over-writeable by sufficiently motivated users.
SPECTRE_EXECUTABLE = "spectre"  # The simulator executable invoked. If over-ridden, likely for sake of a specific path or version.


def available() -> bool:
    """ Boolean indication of whether the current running environment includes the simulator executable. """
    if shutil.which(SPECTRE_EXECUTABLE) is None:
        return False
    try:
        # And if it's set, check that we can get its version without croaking.
        # This can often happen because of an inaccessible license server, or just a badly-linked installation.
        subprocess.run(
            f"{SPECTRE_EXECUTABLE} -V",  # Yes, "version" gets a capital "V" for this program (ascii shrug)
            shell=True,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    except Exception:
        # Indicate "not available" for any Exception. Usually this will be a `subprocess.CalledProcessError`.
        return False
    return True  # Otherwise, installation looks good.


def sim(inp: vsp.SimInput, opts: Optional[SimOptions] = None,) -> SimResultUnion:
    """ # Primary Simulation Method """

    if opts is None:  # Create the default options
        opts = SimOptions(simulator=SupportedSimulators.SPECTRE)

    results = SpectreSim.sim(inp, opts)

    if opts.fmt == ResultFormat.VLSIR_PROTO:
        return results.to_proto()
    return results


class SpectreSim(Sim):
    """
    State and execution logic for a Spectre-call to `vsp.Sim`.
    """

    @classmethod
    def enum(cls) -> SupportedSimulators:
        return SupportedSimulators.SPECTRE

    def _run(self) -> SimResult:
        """ Run the specified `SimInput` in directory `self.tmpdir`, returning its results. """

        netlist_file = open("netlist.scs", "w")
        netlist_file.write("// Test Netlist \n\n")
        netlist_file.write("simulator lang=spectre \n\n")
        netlist_file.write("global 0\n\n")
        netlist(pkg=self.inp.pkg, dest=netlist_file, fmt="spectre")

        # Write the top-level instance
        top_name = SpectreNetlister.get_module_name(self.top)
        netlist_file.write(f"xtop 0 {top_name} // Top-Level DUT \n\n")

        if self.inp.opts:
            raise NotImplementedError(f"SimInput Options")

        # Write each control element
        self.write_control_elements(netlist_file)

        # Write each analysis
        for an in self.inp.an:
            self.netlist_analysis(an, netlist_file)

        # Run the simulation
        netlist_file.flush()
        netlist_file.close()
        self.run_simulation()

        # Parse output data
        data = parse("netlist.raw")
        an_type_dispatch = dict(
            ac=self.parse_ac, dc=self.parse_dc, op=self.parse_op, tran=self.parse_tran
        )
        results = []
        for an in self.inp.an:
            an_type = an.WhichOneof("an")
            inner = getattr(an, an_type)
            if an_type not in an_type_dispatch:
                msg = f"Invalid or Unsupported analysis {an} with type {an_type}"
                raise RuntimeError(msg)
            func = an_type_dispatch[an_type]
            if inner.analysis_name not in data:
                msg = f"Cannot read results for analysis {an}"
                raise RuntimeError(msg)
            inner_data = data[inner.analysis_name]
            an_results = func(inner, inner_data)
            results.append(an_results)

        return SimResult(an=results)

    def write_control_elements(self, netlist_file: IO) -> None:
        """ Write control elements to the netlist """
        for ctrl in self.inp.ctrls:
            inner = ctrl.WhichOneof("ctrl")
            if inner == "include":
                netlist_file.write(f'include "{ctrl.include.path}" \n')
            elif inner == "lib":
                txt = f'include "{ctrl.lib.path}" section={ctrl.lib.section} \n'
                netlist_file.write(txt)
            elif inner == "literal":
                netlist_file.write(ctrl.literal + "\n")
            elif inner == "param":
                # txt = f"parameters {ctrl.param.name}={str(ctrl.param.value)} \n"
                txt = f"parameters  {ctrl.param.name}={SpectreNetlister.get_param_value(ctrl.param.value)} \n"
                netlist_file.write(txt)
            elif inner == "meas":
                # Measurements are written in Spice syntax; wrap them in "simulator lang".
                netlist_file.write(f"simulator lang=spice \n")
                txt = f".meas {ctrl.meas.analysis_type} {ctrl.meas.name} {ctrl.meas.expr} \n"
                netlist_file.write(txt)
                netlist_file.write(f"simulator lang=spectre \n")
            elif inner in ("save"):
                raise NotImplementedError(
                    f"Unimplemented control card {ctrl} for {self}"
                )  # FIXME!
            else:
                raise RuntimeError(f"Unknown control type: {inner}")

    def netlist_analysis(self, an: vsp.Analysis, netlist_file) -> None:
        inner = an.WhichOneof("an")
        inner_dispatch = dict(
            ac=self.netlist_ac,
            dc=self.netlist_dc,
            op=self.netlist_op,
            tran=self.netlist_tran,
        )

        inner_dispatch[inner](getattr(an, inner), netlist_file)

    def netlist_ac(self, an: vsp.AcInput, netlist_file) -> None:
        """ Run an AC analysis. """
        raise NotImplementedError

    def netlist_dc(self, an: vsp.DcInput, netlist_file) -> None:
        """ Run a DC analysis. """
        # Unpack the `DcInput`
        analysis_name = an.analysis_name or "dc"

        if len(an.ctrls):
            raise NotImplementedError  # FIXME!

        # Write the analysis command
        param = an.indep_name
        ## Interpret the sweep
        sweep_type = an.sweep.WhichOneof("tp")
        if sweep_type == "linear":
            sweep = an.sweep.linear
            netlist_file.write(
                f"{analysis_name} dc param={param} start={sweep.start} stop={sweep.stop} step={sweep.step}\n\n"
            )
        elif sweep_type in ("log", "points"):
            raise NotImplementedError
        else:
            raise ValueError("Invalid sweep type")

    def netlist_op(self, an: vsp.OpInput, netlist_file) -> None:
        """
        This netlists as a single point DC analysis
        """
        # Unpack the `OpInput`
        analysis_name = an.analysis_name or "op"
        if len(an.ctrls):
            raise NotImplementedError  # FIXME!

        netlist_file.write(f"{analysis_name} dc oppoint=rawfile\n\n")

    def netlist_tran(self, an: vsp.TranInput, netlist_file) -> None:
        analysis_name = an.analysis_name or "tran"
        if len(an.ctrls):
            raise NotImplementedError
        if len(an.ic):
            raise NotImplementedError

        netlist_file.write(f"{analysis_name} tran stop={an.tstop} \n\n")

    def parse_ac(self, an: vsp.AcInput, data: Mapping[str, Any]) -> vsp.AcResult:
        raise NotImplementedError

    def parse_dc(self, an: vsp.DcInput, data: Mapping[str, Any]) -> vsp.DcResult:
        raise NotImplementedError

    def parse_op(self, an: vsp.OpInput, data: Mapping[str, Any]) -> OpResult:
        return OpResult(
            analysis_name=an.analysis_name,
            data={k: v[0] for k, v in data["data"].items()},
        )

    def parse_tran(self, an: vsp.TranInput, data: Mapping[str, Any]) -> TranResult:
        meas_files = glob.glob("*.mt*")
        if len(meas_files) > 1:
            msg = f"Unsupported: more than one measurement-file generated. Only the first will be read"
            warn(msg)
        elif len(meas_files) == 1:
            measurements = parse_mt0(open(meas_files[0], "r"))
        else:
            measurements = dict()

        return TranResult(
            analysis_name=an.analysis_name, data=data["data"], measurements=measurements
        )

    def run_simulation(self):
        """ Run a Spectre simulation """

        cmd = f"{SPECTRE_EXECUTABLE} -E -format nutbin netlist.scs"
        try:
            subprocess.run(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                shell=True,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise SimProcessError(self, e)
        except Exception as e:
            raise


def parse(filename: str) -> Mapping[str, Any]:
    """ 
    Parse a... what now? PSF Directory? In what format? 
    FIXME: @avi 
    """
    data = {}
    with open(filename, "rb") as f:
        # First 2 lines are ascii one line statements
        _title = f.readline()  # Title, ignored
        _date = f.readline()  # Run date, ignored
        while True:
            # Next 4 lines are also ascii one line statements
            plotname = f.readline()  # Simulation name
            if len(plotname) == 0:
                break
            flags = f.readline()  # Flags for this simulation result
            num_vars_line = f.readline()  # No. Variables:   [nvar]
            num_pts_line = f.readline()  # No. Points:      [npts]
            sim_name = plotname.decode("ascii").split("`")[-1].split("'")[0]
            data[sim_name] = dict(data={}, units={})

            # Find the number of variables and number of points
            num_vars = int(
                re.match(
                    r"No. Variables:\s+(?P<num_vars>\d+)\n",
                    num_vars_line.decode("ascii"),
                ).group("num_vars")
            )
            num_pts = int(
                re.match(
                    r"No. Points:\s+(?P<num_pts>\d+)\n", num_pts_line.decode("ascii"),
                ).group("num_pts")
            )

            # Decode the variables spec, looks like the following
            # Variables: [Variable idx] [Variable name] [units] [optional_flags]
            var_line = f.readline().decode("ascii")
            var_specs = [_read_var_spec(var_line[10:])]
            for i in range(num_vars - 1):
                var_specs.append(_read_var_spec(f.readline().decode("ascii")))

            # Read the binary data, should look like the following:
            # Binary: \n[Binary data]
            binary_line = f.readline().decode("ascii")
            assert binary_line == "Binary:\n"
            # Data is big endian
            bin_data = np.fromfile(
                f, dtype=np.dtype(float).newbyteorder(">"), count=num_vars * num_pts
            )
            for i, var in enumerate(var_specs):
                data[sim_name]["data"][var[0]] = bin_data[i::num_vars]
                data[sim_name]["units"][var[0]] = var[1]

        return data


def _read_var_spec(line: str) -> Tuple[str, str]:
    """Read a Variable spec line from the input
    and return the name and the units."""
    m = re.match(
        r"\s+(?P<idx>\d+)\s+(?P<name>\S+)\s+(?P<units>\S+)(?P<rest>.*)\n", line
    )
    return (m.group("name"), m.group("units"))


def parse_mt0(file: IO) -> Dict[str, float]:
    """ Parse an (open) "mt0-format" measurement-file into a set of {name: value} pairs. """

    file.readline()  # Header
    file.readline()  # Netlist Title
    keys = file.readline()  # Measurement Names Line
    keys = keys.split()
    values = file.readline()  # Measurement Values Line
    values = [float(s) for s in values.split()]
    return dict(zip(keys, values))



"""
# Setup Script

Derived from the setuptools sample project at
https://github.com/pypa/sampleproject/blob/main/setup.py

"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "readme.md").read_text(encoding="utf-8")

_VLSIR_VERSION = "1.0.0rc0"

setup(
    name="vlsirtools",
    version=_VLSIR_VERSION,
    description="Tools for the Vlsir IC Design Schema",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Vlsir/Vlsir",
    author="Dan Fritchman",
    packages=find_packages(),
    python_requires=">=3.7, <4",
    install_requires=[
        f"vlsir=={_VLSIR_VERSION}", 
        "numpy==1.21.5"
    ],
    extras_require={
        "dev": ["pytest==5.2", "coverage", "pytest-cov", "black==19.10b0", "twine"]
    },
)



""" 
# Spice Format Netlisting

"Spice-format" is a bit of a misnomer in netlist-world. 
Of the countless Spice-class simulators have been designed the past half-century, 
most have a similar general netlist format, including: 

* Simulation input comprises a file-full of: 
  * (a) Circuit elements, arranged in `vlsir.Module`s, and
  * (b) Simulator control "cards", such as analysis statements, global parameters, measurements, probes, and the like.
* Circuit-specification is aided by hierarchy, generally in the form of "sub-circuits", denoted `SUBCKT`. 
  * Sub-circuits can commonly be parameterized, and can use a limited set of "parameter programming" to maniupulate their own parameter-values into those of their child-instances. 
  * For example, an instance might be declared as: `xdiode p n area=`width*length``, where `width` and `length` are parameters or its parent.
* `Signal`s are all scalar nets, which are created "out of thin air" whenever referenced. 
* "Typing" performed by instance-name prefixes, e.g. instances named `r1` being interpreted as resistors. 
* Many other subtleties, such as the typical case-insensitivity of netlist content (e.g. `V1` and `v1` are the same net). 

However each simulator also differs in ways large and small. 
Common differences manifest in the areas of: 

* How sub-circuits parameters are declared, and correspondingly how instance-parameter values are set. 
  * Sandia Lab's *Xyce* differs in a prominent fashion, adding a `PARAMS:` keyword where declarations and values begin. 
* How arithmetic expressions are specified, and what functions and expressions are available.
  * Common methods include back-ticks (Hspice) and squiggly-brackets (NgSpice).
  * Notably the asterisk-character (`*`) is the comment-character in many of these formats, and must be wrapped in an expression to perform multiplication. 
* The types and locations of *comments* that are supported. 
  * Some include the fun behavior that comments beginning mid-line require *different* comment-characters from those starting at the beginning of a line.
* While not an HDL attribute, they often differ even more in how simulation control is specified, particularly in analysis and saving is specified. 

"Spice" netlisting therefore requires a small family of "Spice Dialects", 
heavily re-using a central `SpiceNetlister` class, but requiring simulator-specific implementation details. 

"""

# Std-Lib Imports
from typing import Dict, Union

# Local Imports
import vlsir

# Import the base-class
from .base import Netlister, ResolvedModule, ResolvedParams, SpicePrefix, ModuleLike


class SpiceNetlister(Netlister):
    """
    # "Generic" Spice Netlister 
    and base-class for Spice dialects.

    Performs nearly all data-model traversal, 
    offloading syntax-specifics to dialect-specific sub-classes. 

    Attempts to write only the "generic" subset of Spice-content, 
    in the "most generic" methods as perceived by the authors. 
    This may not work for *any* particular simulator; see the simulator-specific dialects below, 
    and the module-level commentary above for more on why. 
    """

    @property
    def enum(self):
        """ Get our entry in the `NetlistFormat` enumeration """
        from . import NetlistFormat

        return NetlistFormat.SPICE

    def write_module_definition(self, module: vlsir.circuit.Module) -> None:
        """ Write the `SUBCKT` definition for `Module` `module`."""

        # Create the module name
        module_name = self.get_module_name(module)
        # Check for double-definition
        if module_name in self.module_names:
            raise RuntimeError(f"Module {module_name} doubly defined")
        # Add to our visited lists
        self.module_names.add(module_name)
        self.pmodules[module.name] = module

        # Create the sub-circuit definition header
        self.write(f".SUBCKT {module_name} \n")

        # Create its ports, if any are defined
        if module.ports:
            self.write_port_declarations(module)
        else:
            self.write("+ ")
            self.write_comment("No ports")

        # Create its parameters, if any are defined
        if module.parameters:
            self.write_param_declarations(module)
        else:
            self.write("+ ")
            self.write_comment("No parameters")

        # End the `subckt` header-content with a blank line
        self.write("\n")

        # Create its instances
        for pinst in module.instances:
            self.write_instance(pinst)

        # And close up the sub-circuit
        self.write(".ENDS\n\n")

    def write_port_declarations(self, module: vlsir.circuit.Module) -> None:
        """ Write the port declarations for Module `module`. """
        self.write("+ ")
        for pport in module.ports:
            self.write(self.format_port_decl(pport) + " ")
        self.write("\n")

    def write_param_declarations(self, module: vlsir.circuit.Module) -> None:
        """ Write the parameter declarations for Module `module`. 
        Parameter declaration format: `name1=val1 name2=val2 name3=val3 \n`"""
        self.write("+ ")
        for name, pparam in module.parameters.items():
            self.write(self.format_param_decl(name, pparam))
        self.write("\n")

    def write_instance_name(
        self, pinst: vlsir.circuit.Instance, rmodule: ResolvedModule
    ) -> None:
        """ Write the instance-name line for `pinst`, including the SPICE-dictated primitive-prefix. """
        self.write(f"{rmodule.spice_prefix.value}{pinst.name} \n")

    def write_instance(self, pinst: vlsir.circuit.Instance) -> None:
        """ Create and return a netlist-string for Instance `pinst`"""

        # Get its Module or ExternalModule definition,
        resolved = self.resolve_reference(pinst.module)

        # And dispatch to `subckt` or `primitive` writers
        if resolved.spice_prefix == SpicePrefix.SUBCKT:
            return self.write_subckt_instance(pinst, resolved)

        if resolved.spice_prefix == SpicePrefix.VSOURCE:
            # Voltage sources get weird, and vary between dialiects. Farm them out to a dedicated method.
            return self.write_voltage_source_instance(pinst, resolved)

        # Everything else falls into the `primitive` category
        return self.write_primitive_instance(pinst, resolved)

    def write_subckt_instance(
        self, pinst: vlsir.circuit.Instance, rmodule: ResolvedModule
    ) -> None:
        """ Write sub-circuit-instance `pinst` of `rmodule`. """

        # Write the instance name
        self.write_instance_name(pinst, rmodule)

        # Write its port-connections
        self.write_instance_conns(pinst, rmodule.module)

        # Write the sub-circuit name
        self.write("+ " + rmodule.module_name + " \n")

        # Write its parameter values
        resolved_param_values = self.get_instance_params(pinst, rmodule.module)
        self.write_instance_params(resolved_param_values)

        # Add a blank after each instance
        self.write("\n")

    def write_primitive_instance(
        self, pinst: vlsir.circuit.Instance, rmodule: ResolvedModule
    ) -> None:
        """ Write primitive-instance `pinst` of `rmodule`. 
        Note spice's primitive instances often differn syntactically from sub-circuit instances, 
        in that they can have positional (only) parameters. """

        # Write the instance name
        self.write_instance_name(pinst, rmodule)

        # Write its port-connections
        self.write_instance_conns(pinst, rmodule.module)

        # Resolve its parameter-values to spice-strings
        resolved_param_values = self.get_instance_params(pinst, rmodule.module)

        # Write special and/or positional parameters
        if rmodule.spice_prefix == SpicePrefix.RESISTOR:
            positional_keys = ["r"]
        elif rmodule.spice_prefix == SpicePrefix.CAPACITOR:
            positional_keys = ["c"]
        elif rmodule.spice_prefix == SpicePrefix.INDUCTOR:
            positional_keys = ["l"]
        elif rmodule.spice_prefix == SpicePrefix.ISOURCE:
            positional_keys = ["dc"]
        elif rmodule.spice_prefix in (
            SpicePrefix.VCVS,
            SpicePrefix.VCCS,
            SpicePrefix.CCCS,
            SpicePrefix.CCVS,
        ):
            positional_keys = ["gain"]
        elif rmodule.spice_prefix in (
            SpicePrefix.MOS,
            SpicePrefix.BIPOLAR,
            SpicePrefix.DIODE,
            SpicePrefix.TLINE,
        ):
            positional_keys = ["modelname"]
        else:
            positional_keys = []

        # Pop all positional parameters ("pp") from `resolved_param_values`
        pp = resolved_param_values.pop_many(positional_keys)

        # Write the positional parameters, in the order specified by `positional_keys`
        self.write("+ " + " ".join([pp[pkey] for pkey in positional_keys]) + " \n")

        # Now! Write its subckt-style by-name parameter values
        self.write_instance_params(resolved_param_values)

        # Add a blank after each instance
        self.write("\n")

    def write_voltage_source_instance(
        self, pinst: vlsir.circuit.Instance, rmodule: ResolvedModule,
    ) -> None:
        """ Write a voltage-source instance `pinst`.
        Throws an Exception if `rmodule` is not a known voltage-source type. """

        # Resolve its parameter values
        resolved_param_values = self.get_instance_params(pinst, rmodule.module)

        # Write the instance name
        self.write_instance_name(pinst, rmodule)

        # Write its port-connections
        self.write_instance_conns(pinst, rmodule.module)

        # Handle each of the voltage-source cases
        name = rmodule.module.name.name
        if name == "vdc":
            dc = resolved_param_values.pop("dc")
            self.write(f"+ dc {self.format_expression(dc)} \n")

        elif name == "vpulse":
            keys = ["v1", "v2", "td", "tr", "tf", "tpw", "tper"]
            pp = resolved_param_values.pop_many(keys)
            self.write(f"+ pulse (" + " ".join([self.format_expression(pp[k]) for k in keys]) + ") \n")

        elif name == "vsin":
            keys = ["voff", "vamp", "freq", "td", "phase"]
            pp = resolved_param_values.pop_many(keys)
            self.write(f"+ sin (" + " ".join([self.format_expression(pp[k]) for k in keys]) + ") \n")

        else:
            raise ValueError(f"Invalid or unsupported voltage-source type: {name}")

        # Now! Write its subckt-style by-name parameter values
        self.write_instance_params(resolved_param_values)

        # Add a blank after each instance
        self.write("\n")

    def write_instance_conns(
        self, pinst: vlsir.circuit.Instance, module: ModuleLike
    ) -> None:
        """ Write the port-connections for Instance `pinst` """

        # Write a quick comment for port-less modules
        if not module.ports:
            self.write("+ ")
            return self.write_comment("No ports")

        self.write("+ ")
        # And write the Instance ports, in that order
        port_order = [pport.signal.name for pport in module.ports]
        for pname in port_order:
            pconn = pinst.connections.get(pname, None)
            if pconn is None:
                raise RuntimeError(f"Unconnected Port {pname} on {pinst.name}")
            self.write(self.format_connection(pconn) + " ")
        self.write("\n")

    def write_instance_params(self, pvals: ResolvedParams) -> None:
        """ 
        Format and write the parameter-values in dictionary `pvals`. 
        
        Parameter-values format:
        ```
        XNAME 
        + <ports> 
        + <subckt-name> 
        + name1=val1 name2=val2 name3=val3 
        """

        self.write("+ ")

        if not pvals:  # Write a quick comment for no parameters
            self.write_comment("No parameters")

        # And write them
        for (pname, pval) in pvals.items():
            self.write(f"{pname}={self.format_expression(pval)} ")

        self.write("\n")

    def format_concat(self, pconc: vlsir.circuit.Concat) -> str:
        """ Format the Concatenation of several other Connections """
        out = ""
        for part in pconc.parts:
            out += self.format_connection(part) + " "
        return out

    @classmethod
    def format_port_decl(cls, pport: vlsir.circuit.Port) -> str:
        """ Get a netlist `Port` definition """
        return cls.format_signal_ref(pport.signal)

    @classmethod
    def format_port_ref(cls, pport: vlsir.circuit.Port) -> str:
        """ Get a netlist `Port` reference """
        return cls.format_signal_ref(pport.signal)

    @classmethod
    def format_signal_ref(cls, psig: vlsir.circuit.Signal) -> str:
        """ Get a netlist definition for Signal `psig` """
        if psig.width < 1:
            raise RuntimeError
        if psig.width == 1:  # width==1, i.e. a scalar signal
            return psig.name
        # Vector/ multi "bit" Signal. Creates several spice signals.
        return " ".join(
            [f"{psig.name}{cls.format_bus_bit(k)}" for k in reversed(range(psig.width))]
        )

    @classmethod
    def format_signal_slice(cls, pslice: vlsir.circuit.Slice) -> str:
        """ Get a netlist definition for Signal-Slice `pslice` """
        base = pslice.signal
        indices = list(reversed(range(pslice.bot, pslice.top + 1)))
        if not len(indices):
            raise RuntimeError(f"Attempting to netlist empty slice {pslice}")
        return " ".join([f"{base}{cls.format_bus_bit(k)}" for k in indices])

    @classmethod
    def format_bus_bit(cls, index: Union[int, str]) -> str:
        """ Format-specific string-representation of a bus bit-index """
        # Spectre netlisting uses an underscore prefix, e.g. `bus_0`
        return "_" + str(index)

    @classmethod
    def format_param_decl(cls, name: str, param: vlsir.Param) -> str:
        """ Format a parameter-declaration """
        default = cls.get_param_default(param)
        if default is None:
            msg = f"Invalid non-default parameter {param} for Spice netlisting"
            raise RuntimeError(msg)
        return f"{name}={default}"

    def write_comment(self, comment: str) -> None:
        """ While dialects vary, the *generic* Spice-comment begins with the asterisk. """
        self.write(f"* {comment}\n")

    def format_expression(self, expr: str) -> str:
        """ Format a string such that the target format interprets it as an expression. 
        Example:
        ```
        * Parameter Declarations
        .param v0=1 v1='v0+1' * <= Here
        * Instance with the same name 
        v0 1 0 dc='v0+2*v1' * <= And here 
        ``` 
        Note the latter case includes the star character (`*`) for multiplication, 
        where in many other contexts it is treated as the comment-character. 
        """
        # The base class does what (we think) is the most common practice:
        # wrapping expressions in single-tick quotes.
        return f"'{expr}'"


class HspiceNetlister(SpiceNetlister):
    """
    # Hspice-Format Netlister 
    
    Other than its `NetlistFormat` enumeration, `HspiceNetlister` is identical to the base `SpiceNetlister`. 
    """

    @property
    def enum(self):
        """ Get our entry in the `NetlistFormat` enumeration """
        from . import NetlistFormat

        return NetlistFormat.HSPICE


class XyceNetlister(SpiceNetlister):
    """ Xyce-Format Netlister """

    @property
    def enum(self):
        """ Get our entry in the `NetlistFormat` enumeration """
        from . import NetlistFormat

        return NetlistFormat.XYCE

    def write_param_declarations(self, module: vlsir.circuit.Module) -> None:
        """ Write the parameter declarations for Module `module`. 
        Parameter declaration format:
        .SUBCKT <name> <ports> 
        + PARAMS: name1=val1 name2=val2 name3=val3 \n
        """
        self.write("+ PARAMS: ")  # <= Xyce-specific
        for name, pparam in module.parameters.items():
            self.write(self.format_param_decl(name, pparam))
        self.write("\n")

    def write_instance_params(self, pvals: Dict[str, str]) -> None:
        """ Write the parameter-values for Instance `pinst`. 

        Parameter-values format:
        ```
        XNAME 
        + <ports> 
        + <subckt-name> 
        + PARAMS: name1=val1 name2=val2 name3=val3 
        """

        self.write("+ ")
        if not pvals:  # Write a quick comment for no parameters
            return self.write_comment("No parameters")

        self.write("PARAMS: ")  # <= Xyce-specific
        for (pname, pval) in pvals.items():
            self.write(f"{pname}={self.format_expression(pval)} ")

        self.write("\n")

    def write_comment(self, comment: str) -> None:
        """ Xyce comments *kinda* support the Spice-typical `*` charater, 
        but *only* as the first character in a line. 
        Any mid-line-starting comments must use `;` instead. 
        So, just use it all the time. """
        self.write(f"; {comment}\n")

    def format_expression(self, expr: str) -> str:
        # Xyce expressions are wrapped in curly braces.
        return f"{{{expr}}}"


class NgspiceNetlister(SpiceNetlister):
    """ FIXME: Ngspice-Format Netlister """

    def __init__(self, *_, **__):
        raise NotImplementedError

    @property
    def enum(self):
        """ Get our entry in the `NetlistFormat` enumeration """
        from . import NetlistFormat

        return NetlistFormat.NGSPICE


class CdlNetlister(SpiceNetlister):
    """ FIXME: CDL-Format Netlister """

    def __init__(self, *_, **__):
        raise NotImplementedError

    @property
    def enum(self):
        """ Get our entry in the `NetlistFormat` enumeration """
        from . import NetlistFormat

        return NetlistFormat.CDL
